package ru.dbotthepony.kstarbound.client.gl.shader

import it.unimi.dsi.fastutil.objects.Object2BooleanArrayMap
import it.unimi.dsi.fastutil.objects.Object2BooleanFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.checkForGLError
import ru.dbotthepony.kvector.api.IFloatMatrix
import ru.dbotthepony.kvector.api.IStruct2f
import ru.dbotthepony.kvector.api.IStruct3f
import ru.dbotthepony.kvector.api.IStruct4f
import ru.dbotthepony.kvector.api.concrete.IMatrix3f
import ru.dbotthepony.kvector.api.concrete.IMatrix4f
import ru.dbotthepony.kvector.matrix.nfloat.Matrix3f
import ru.dbotthepony.kvector.matrix.nfloat.Matrix4f
import ru.dbotthepony.kvector.vector.nfloat.Vector2f
import ru.dbotthepony.kvector.vector.nfloat.Vector3f
import ru.dbotthepony.kvector.vector.nfloat.Vector4f
import java.nio.ByteBuffer
import java.nio.ByteOrder
import java.nio.FloatBuffer
import java.util.*
import java.util.stream.Stream
import kotlin.NoSuchElementException
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KProperty

open class GLShaderProgram(val state: GLStateTracker, shaders: Iterable<GLStateTracker.Shader>) {
	init {
		state.ensureSameThread()
	}

	val pointer = glCreateProgram()

	init {
		checkForGLError("Creating shader program")
		state.registerCleanable(this, ::glDeleteProgram, pointer)

		for (shader in shaders) {
			glAttachShader(pointer, shader.pointer)
			checkForGLError("Attaching shader")
		}

		glLinkProgram(pointer)

		val success = intArrayOf(0)
		glGetProgramiv(pointer, GL_LINK_STATUS, success)

		if (success[0] == 0) {
			throw ShaderLinkException(glGetProgramInfoLog(pointer))
		}

		glGetError()
	}

	fun use() = state.use(this)

	private val locations = Object2ObjectArrayMap<String, Uniform<*>>()
	private val uniformsExist = Object2BooleanArrayMap<String>()

	fun isUniformPresent(name: String): Boolean {
		state.ensureSameThread()
		return uniformsExist.computeIfAbsent(name, Object2BooleanFunction { glGetUniformLocation(pointer, name) != -1 })
	}

	fun getUniform(name: String, orElse: ((String) -> Uniform<*>)? = null): Uniform<*>? {
		return locations[name] ?: orElse?.invoke(name)
	}

	fun int(name: String): IUniform {
		return locations.computeIfAbsent(name, Object2ObjectFunction { IUniform(name) }) as? IUniform ?: throw IllegalStateException("Uniform $name has type of ${locations[name]!!::class.simpleName}")
	}

	abstract inner class Uniform<V : Any>(val name: String) : ReadWriteProperty<Any?, V> {
		init {
			state.ensureSameThread()
			require(!locations.containsKey(name)) { "Already has uniform $name for ${this@GLShaderProgram} (${locations[name]})" }
		}

		val location = glGetUniformLocation(pointer, name)

		init {
			if (location == -1)
				throw NoSuchElementException("Program ${this@GLShaderProgram} does not have uniform with name $name")

			locations[name] = this
		}

		abstract var value: V

		override fun toString(): String {
			return "Uniform[$name, $value]"
		}

		// они переопределены в дочерних классах для бетонной реализации примитивов
		override fun getValue(thisRef: Any?, property: KProperty<*>): V {
			return value
		}

		override fun setValue(thisRef: Any?, property: KProperty<*>, value: V) {
			this.value = value
		}
	}

	inner class FUniform(name: String) : Uniform<Float>(name) {
		override var value: Float = 0f
			set(value) {
				state.ensureSameThread()

				if (field != value) {
					glProgramUniform1f(pointer, location, value)
					checkForGLError()
				}

				field = value
			}

		override fun getValue(thisRef: Any?, property: KProperty<*>): Float {
			return value
		}

		override fun setValue(thisRef: Any?, property: KProperty<*>, value: Float) {
			this.value = value
		}
	}

	inner class F2Uniform(name: String) : Uniform<IStruct2f>(name) {
		override var value: IStruct2f = Vector2f.ZERO
			set(value) {
				state.ensureSameThread()

				if (field != value) {
					glProgramUniform2f(pointer, location, value.component1(), value.component2())
					checkForGLError()

					field = if (value is Vector2f) value else Vector2f(value)
				}
			}
	}

	inner class F3Uniform(name: String) : Uniform<IStruct3f>(name) {
		override var value: IStruct3f = Vector3f.ZERO
			set(value) {
				state.ensureSameThread()

				if (field != value) {
					glProgramUniform3f(pointer, location, value.component1(), value.component2(), value.component3())
					checkForGLError()

					field = if (value is Vector3f) value else Vector3f(value)
				}
			}
	}

	private val buff3x3: FloatBuffer by lazy(LazyThreadSafetyMode.NONE) { ByteBuffer.allocateDirect(4 * 3 * 3).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer() }
	private val buff4x4: FloatBuffer by lazy(LazyThreadSafetyMode.NONE) { ByteBuffer.allocateDirect(4 * 4 * 4).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer() }

	inner class F3x3Uniform(name: String) : Uniform<IMatrix3f<*>>(name) {
		override var value: IMatrix3f<*> = Matrix3f.ZERO
			set(value) {
				state.ensureSameThread()

				if (field != value) {
					buff3x3.position(0)
					value.write(buff3x3)
					buff3x3.position(0)
					glProgramUniformMatrix3fv(pointer, location, false, buff3x3)
					checkForGLError()

					field = value.toMatrix3f()
				}
			}
	}

	inner class F4x4Uniform(name: String) : Uniform<IMatrix4f<*>>(name) {
		override var value: IMatrix4f<*> = Matrix4f.ZERO
			set(value) {
				state.ensureSameThread()

				if (field != value) {
					buff4x4.position(0)
					value.write(buff4x4)
					buff4x4.position(0)
					glProgramUniformMatrix4fv(pointer, location, false, buff4x4)
					checkForGLError()

					field = value.toMatrix4f()
				}
			}
	}

	inner class F4Uniform(name: String) : Uniform<IStruct4f>(name) {
		override var value: IStruct4f = Vector4f.ZERO
			set(value) {
				state.ensureSameThread()

				if (field != value) {
					glProgramUniform4f(pointer, location, value.component1(), value.component2(), value.component3(), value.component4())
					checkForGLError()

					field = if (value is Vector4f) value else Vector4f(value)
				}
			}
	}

	inner class IUniform(name: String) : Uniform<Int>(name) {
		override var value: Int = 0
			set(value) {
				state.ensureSameThread()

				if (field != value) {
					glProgramUniform1i(pointer, location, value)
					checkForGLError()
				}

				field = value
			}

		override fun getValue(thisRef: Any?, property: KProperty<*>): Int {
			return value
		}

		override fun setValue(thisRef: Any?, property: KProperty<*>, value: Int) {
			this.value = value
		}
	}
}
